import logging
import os
from typing import List

from quollio_data_profiler.core import setup
from quollio_data_profiler.profilers.stats_profiler import gen_table_stats_payload
from quollio_data_profiler.repository import dbt, qdc, snowflake

logger = logging.getLogger(__name__)


def _build_query(target_databases: list) -> str:
    if len(target_databases) == 0:
        query = """
            SELECT
                DISTINCT
                TABLE_CATALOG
                , TABLE_SCHEMA
                , TABLE_NAME
            FROM
                QUOLLIO_DATA_PROFILER.PUBLIC.PROFILE_TARGET_COLUMNS
            """
    else:
        db_str = ", ".join("'" + db + "'" for db in target_databases)
        query = """
            SELECT
                DISTINCT
                TABLE_CATALOG
                , TABLE_SCHEMA
                , TABLE_NAME
            FROM
                QUOLLIO_DATA_PROFILER.PUBLIC.PROFILE_TARGET_COLUMNS
            WHERE
                TABLE_CATALOG in({databases})
            """.format(
            databases=db_str
        )
    return query


def snowflake_table_stats(
    company_id: str,
    snowflake_connections: snowflake.SnowflakeConnectionConfig,
    qdc_client: qdc.QDCExternalAPIClient,
    target_databases: list = [],
) -> None:
    logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(name)s - %(message)s")
    sf_executor = snowflake.SnowflakeQueryExecutor(snowflake_connections, "QUOLLIO_DATA_PROFILER", "PUBLIC")

    query = _build_query(target_databases=target_databases)
    target_assets = sf_executor.get_query_results(query=query)
    req_count = 0
    for target_asset in target_assets:
        stats_query = """
        SELECT
            *
        FROM
            QUOLLIO_DATA_PROFILER.PUBLIC.COLUMN_STATISTICS
        WHERE
            DB_NAME = '{db}'
            AND SCHEMA_NAME = '{schema}'
            AND TABLE_NAME = '{table}'
            AND DB_NAME != 'QUOLLIO_DATA_PROFILER'
        """.format(
            db=target_asset["TABLE_CATALOG"], schema=target_asset["TABLE_SCHEMA"], table=target_asset["TABLE_NAME"]
        )
        stats_result = sf_executor.get_query_results(query=stats_query)
        payloads = gen_table_stats_payload(
            company_id=company_id, endpoint=snowflake_connections.account_id, stats=stats_result
        )
        for payload in payloads:
            logger.info(
                "Generating table stats. asset: {db} -> {schema} -> {table} -> {column}".format(
                    db=payload.db,
                    schema=payload.schema,
                    table=payload.table,
                    column=payload.column,
                )
            )
            status_code = qdc_client.update_stats_by_id(
                global_id=payload.global_id,
                payload=payload.body.get_column_stats(),
            )
            if status_code == 200:
                req_count += 1
    logger.info(f"Generating table stats is finished. {req_count} stats are ingested.")


def execute(
    company_id: str,
    sf_build_view_connections: snowflake.SnowflakeConnectionConfig,
    qdc_client: qdc.QDCExternalAPIClient,
    is_view_build_only: bool,
    sf_query_connections: snowflake.SnowflakeConnectionConfig,
    stats_target_databases: List[str] = [],
    stats_sample_method: str = "SAMPLE(10)",
) -> None:
    logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(name)s - %(message)s")

    logger.info("Create stats view")
    dbt_client = dbt.DBTClient()

    current_dir = os.path.dirname(os.path.abspath(__file__))
    project_path = f"{current_dir}/profilers/dbt_projects/snowflake_stats_profiler"
    template_path = f"{current_dir}/profilers/templates"
    setup.setup_dbt_profile(
        connections=sf_build_view_connections,
        project_path=project_path,
        template_path=template_path,
        template_name="snowflake_stats_project.j2",
    )
    var_settings = '{{"exec_role": {exec_role}, "sample_method": {sample_method}}}'.format(
        exec_role=sf_query_connections.account_role, sample_method=stats_sample_method
    )

    # FIXME: when executing some of the commands, directory changes due to the library bug.
    # https://github.com/dbt-labs/dbt-core/issues/8997
    dbt_client.invoke(cmd="deps", project_dir=project_path, profile_dir=project_path, options=["--vars", var_settings])
    dbt_client.invoke(cmd="run", project_dir=".", profile_dir=".", options=["--vars", var_settings])

    if is_view_build_only:
        logger.info("Skip ingesting metadata into QDC.")
        return

    snowflake_table_stats(
        company_id=company_id,
        snowflake_connections=sf_query_connections,
        qdc_client=qdc_client,
        target_databases=stats_target_databases,
    )



from dataclasses import asdict, dataclass
from decimal import ROUND_HALF_UP, Decimal
from typing import Dict, List

from quollio_data_profiler.core.core import new_global_id


@dataclass
class ColumnStatsInput:
    cardinality: int
    max: str
    mean: str
    median: str
    min: str
    mode: str
    number_of_null: int
    number_of_unique: int
    stddev: str

    def as_dict(self) -> Dict[str, str]:
        return asdict(self)


@dataclass
class TableStatsInput:
    count: int
    size: float

    def as_dict(self) -> Dict[str, str]:
        return asdict(self)


@dataclass
class StatsInput:
    column_stats: ColumnStatsInput
    table_stats: TableStatsInput

    def as_dict(self) -> Dict[str, str]:
        return asdict(self)

    def get_column_stats(self):
        return {"column_stats": self.column_stats.as_dict()}


@dataclass
class StatsRequest:
    global_id: str
    db: str
    schema: str
    table: str
    column: str
    body: StatsInput

    def as_dict(self) -> Dict[str, str]:
        return asdict(self)


def convert_value_type(obj):
    if isinstance(obj, Decimal):
        return str(obj.quantize(Decimal("0.1"), ROUND_HALF_UP))
    return obj


def gen_table_stats_payload(company_id: str, endpoint: str, stats: List[Dict[str, str]]) -> List[StatsRequest]:
    payloads = list()
    for stat in stats:
        global_id_arg = "{db}{schema}{table}{column}".format(
            db=stat["DB_NAME"], schema=stat["SCHEMA_NAME"], table=stat["TABLE_NAME"], column=stat["COLUMN_NAME"]
        )
        table_global_id = new_global_id(
            company_id=company_id, cluster_id=endpoint, data_id=global_id_arg, data_type="column"
        )

        stats_request = StatsRequest(
            global_id=table_global_id,
            db=stat["DB_NAME"],
            schema=stat["SCHEMA_NAME"],
            table=stat["TABLE_NAME"],
            column=stat["COLUMN_NAME"],
            body=StatsInput(
                column_stats=ColumnStatsInput(
                    cardinality=convert_value_type(stat["CARDINALITY"]),
                    max=str(convert_value_type(stat["MAX_VALUE"])),
                    mean=str(convert_value_type(stat["AVG_VALUE"])),
                    median=str(convert_value_type(stat["MEDIAN_VALUE"])),
                    min=str(convert_value_type(stat["MIN_VALUE"])),
                    mode=str(convert_value_type(stat["MODE_VALUE"])),
                    number_of_null=convert_value_type(stat["NULL_COUNT"]),
                    number_of_unique=convert_value_type(stat["CARDINALITY"]),
                    stddev=str(convert_value_type(stat["STDDEV_VALUE"])),
                ),
                # MEMO: Table stats can be collected with metadata agent.
                # Then, It's not necessary to update with this system for now.
                table_stats=TableStatsInput(count=0, size=0.0),
            ),
        )
        payloads.append(stats_request)
    return payloads



import json
from dataclasses import asdict, dataclass
from typing import Dict, List

from quollio_data_profiler.core.core import new_global_id


@dataclass
class LineageInput:
    upstream: List[str]

    def as_dict(self) -> Dict[str, str]:
        return asdict(self)


@dataclass
class LineageInputs:
    downstream_global_id: str
    downstream_database_name: str
    downstream_schema_name: str
    downstream_table_name: str
    downstream_column_name: str
    upstreams: LineageInput


def gen_table_lineage_payload(company_id: str, endpoint: str, tables: List[Dict[str, str]]) -> List[LineageInputs]:
    payload = list()
    for table in tables:
        downstream_table_fqdn = table["DOWNSTREAM_TABLE_NAME"].split(".")
        if len(downstream_table_fqdn) != 3:
            continue
        else:
            global_id_arg = "{db}{schema}{table}".format(
                db=downstream_table_fqdn[0], schema=downstream_table_fqdn[1], table=downstream_table_fqdn[2]
            )
            downstream_table_global_id = new_global_id(
                company_id=company_id, cluster_id=endpoint, data_id=global_id_arg, data_type="table"
            )
            upstream_tables: List[Dict[str, str]] = json.loads(table["UPSTREAM_TABLES"])
            lineage_input = LineageInput(upstream=[])
            for upstream_table in upstream_tables:
                upstream_table_fqdn = upstream_table["upstream_object_name"].split(".")
                if len(upstream_table_fqdn) != 3:
                    continue
                else:
                    upstream_global_id_arg = "{db}{schema}{table}".format(
                        db=upstream_table_fqdn[0], schema=upstream_table_fqdn[1], table=upstream_table_fqdn[2]
                    )
                    upstream_table_global_id = new_global_id(
                        company_id=company_id, cluster_id=endpoint, data_id=upstream_global_id_arg, data_type="table"
                    )
                    lineage_input.upstream.append(upstream_table_global_id)
            lineage_inputs = LineageInputs(
                downstream_global_id=downstream_table_global_id,
                downstream_database_name=downstream_table_fqdn[0],
                downstream_schema_name=downstream_table_fqdn[1],
                downstream_table_name=downstream_table_fqdn[2],
                downstream_column_name="",
                upstreams=lineage_input,
            )
            payload.append(lineage_inputs)
    return payload


def gen_column_lineage_payload(company_id: str, endpoint: str, columns: List[Dict[str, str]]) -> List[LineageInputs]:
    payload = list()
    for column in columns:
        downstream_table_fqdn = column["DOWNSTREAM_TABLE_NAME"].split(".")
        if len(downstream_table_fqdn) != 3:
            continue
        else:
            global_id_arg = "{db}{schema}{table}{column}".format(
                db=downstream_table_fqdn[0],
                schema=downstream_table_fqdn[1],
                table=downstream_table_fqdn[2],
                column=column["DOWNSTREAM_COLUMN_NAME"],
            )
            downstream_column_global_id = new_global_id(
                company_id=company_id, cluster_id=endpoint, data_id=global_id_arg, data_type="column"
            )
            upstream_columns: List[Dict[str, str]] = json.loads(column["UPSTREAM_COLUMNS"])
            lineage_input = LineageInput(upstream=[])
            for upstream_column in upstream_columns:
                upstream_table_fqdn = upstream_column["upstream_table_name"].split(".")
                if len(upstream_table_fqdn) != 3:
                    continue
                elif not upstream_column.get("upstream_column_name"):
                    continue
                else:
                    upstream_global_id_arg = "{db}{schema}{table}{column}".format(
                        db=upstream_table_fqdn[0],
                        schema=upstream_table_fqdn[1],
                        table=upstream_table_fqdn[2],
                        column=upstream_column["upstream_column_name"],
                    )
                    upstream_column_global_id = new_global_id(
                        company_id=company_id, cluster_id=endpoint, data_id=upstream_global_id_arg, data_type="column"
                    )
                    lineage_input.upstream.append(upstream_column_global_id)
            lineage_inputs = LineageInputs(
                downstream_global_id=downstream_column_global_id,
                downstream_database_name=downstream_table_fqdn[0],
                downstream_schema_name=downstream_table_fqdn[1],
                downstream_table_name=downstream_table_fqdn[2],
                downstream_column_name=column["DOWNSTREAM_COLUMN_NAME"],
                upstreams=lineage_input,
            )
            payload.append(lineage_inputs)
    return payload



"""Quollio Data Profiler"""

__version__ = "0.2.1"
__author__ = "Quollio technologies"


from quollio_data_profiler.repository.snowflake import SnowflakeConnectionConfig  # noqa:F401
from quollio_data_profiler.snowflake_lineage_profiler import execute  # noqa:F401



import logging
import os

from quollio_data_profiler.core import setup
from quollio_data_profiler.profilers.lineage_profiler import gen_column_lineage_payload, gen_table_lineage_payload
from quollio_data_profiler.repository import dbt, qdc, snowflake

logger = logging.getLogger(__name__)


def snowflake_table_to_table_lineage(
    company_id: str, snowflake_connections: snowflake.SnowflakeConnectionConfig, qdc_client: qdc.QDCExternalAPIClient
) -> None:
    logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(name)s - %(message)s")
    sf_executor = snowflake.SnowflakeQueryExecutor(snowflake_connections, "QUOLLIO_DATA_PROFILER", "PUBLIC")
    results = sf_executor.get_query_results(
        query="""
        SELECT
            *
        FROM
            QUOLLIO_DATA_PROFILER.PUBLIC.TABLE_TO_TABLE_LINEAGE
        """
    )
    update_table_lineage_inputs = gen_table_lineage_payload(
        company_id=company_id,
        endpoint=snowflake_connections.account_id,
        tables=results,
    )

    req_count = 0
    for update_table_lineage_input in update_table_lineage_inputs:
        logger.info(
            "Generating table lineage. downstream: {db} -> {schema} -> {table}".format(
                db=update_table_lineage_input.downstream_database_name,
                schema=update_table_lineage_input.downstream_schema_name,
                table=update_table_lineage_input.downstream_table_name,
            )
        )
        status_code = qdc_client.update_lineage_by_id(
            global_id=update_table_lineage_input.downstream_global_id,
            payload=update_table_lineage_input.upstreams.as_dict(),
        )
        if status_code == 200:
            req_count += 1
    logger.info(f"Generating table lineage is finished. {req_count} lineages are ingested.")
    return


def snowflake_column_to_column_lineage(
    company_id: str, snowflake_connections: snowflake.SnowflakeConnectionConfig, qdc_client: qdc.QDCExternalAPIClient
) -> None:
    sf_executor = snowflake.SnowflakeQueryExecutor(snowflake_connections, "QUOLLIO_DATA_PROFILER", "PUBLIC")
    results = sf_executor.get_query_results(
        query="""
        SELECT
            *
        FROM
            QUOLLIO_DATA_PROFILER.PUBLIC.COLUMN_TO_COLUMN_LINEAGE
        """
    )
    update_column_lineage_inputs = gen_column_lineage_payload(
        company_id=company_id,
        endpoint=snowflake_connections.account_id,
        columns=results,
    )

    req_count = 0
    for update_column_lineage_input in update_column_lineage_inputs:
        logger.info(
            "Generating column lineage. downstream: {db} -> {schema} -> {table} -> {column}".format(
                db=update_column_lineage_input.downstream_database_name,
                schema=update_column_lineage_input.downstream_schema_name,
                table=update_column_lineage_input.downstream_table_name,
                column=update_column_lineage_input.downstream_column_name,
            )
        )
        status_code = qdc_client.update_lineage_by_id(
            global_id=update_column_lineage_input.downstream_global_id,
            payload=update_column_lineage_input.upstreams.as_dict(),
        )
        if status_code == 200:
            req_count += 1
    logger.info(f"Generating column lineage is finished. {req_count} lineages are ingested.")
    return


def execute(
    company_id: str,
    sf_build_view_connections: snowflake.SnowflakeConnectionConfig,
    qdc_client: qdc.QDCExternalAPIClient,
    is_view_build_only: bool,
    sf_query_connections: snowflake.SnowflakeConnectionConfig,
) -> None:
    logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(name)s - %(message)s")

    logger.info("Create lineage view")
    dbt_client = dbt.DBTClient()

    current_dir = os.path.dirname(os.path.abspath(__file__))
    project_path = f"{current_dir}/profilers/dbt_projects/snowflake_lineage_profiler"
    template_path = f"{current_dir}/profilers/templates"
    setup.setup_dbt_profile(
        connections=sf_build_view_connections,
        project_path=project_path,
        template_path=template_path,
        template_name="snowflake_lineage_project.j2",
    )
    exec_role_settings = '{{"exec_role": {exec_role}}}'.format(exec_role=sf_query_connections.account_role)

    # FIXME: when executing some of the commands, directory changes due to the library bug.
    # https://github.com/dbt-labs/dbt-core/issues/8997
    dbt_client.invoke(
        cmd="deps", project_dir=project_path, profile_dir=project_path, options=["--vars", exec_role_settings]
    )

    dbt_client.invoke(cmd="run", project_dir=".", profile_dir=".", options=["--vars", exec_role_settings])

    if is_view_build_only:
        logger.info("Skip ingesting metadata into QDC.")
        return

    logger.info("Generate snowflake table to table lineage.")
    snowflake_table_to_table_lineage(
        company_id=company_id, snowflake_connections=sf_query_connections, qdc_client=qdc_client
    )

    logger.info("Generate snowflake column to column lineage.")
    snowflake_column_to_column_lineage(
        company_id=company_id, snowflake_connections=sf_query_connections, qdc_client=qdc_client
    )

    logger.info("Snowflake lineage profiler is successfully finished.")
    return



from blake3 import blake3


def new_global_id(company_id: str, cluster_id: str, data_id: str, data_type: str) -> str:
    prefix = ""
    data_types = {
        "schema": "schm-",
        "table": "tbl-",
        "column": "clmn-",
        "bigroup": "bgrp-",
        "dashboard": "dsbd-",
        "sheet": "sht-",
    }
    prefix = data_types[data_type]

    data_to_hash = f"{company_id}{cluster_id}{data_id}"
    hashed = blake3(data_to_hash.encode()).digest()
    global_id: str = prefix + hashed[:16].hex()
    return global_id



import yaml
from jinja2 import Environment, FileSystemLoader

from quollio_data_profiler.repository.snowflake import SnowflakeConnectionConfig


def setup_dbt_profile(
    connections: SnowflakeConnectionConfig, project_path: str, template_path: str, template_name: str
) -> None:
    connections_json = connections.as_dict()

    profile_path = f"{project_path}/profiles.yml"
    loader = Environment(loader=(FileSystemLoader(template_path, encoding="utf-8")))
    template = loader.get_template(template_name)
    profiles_body = template.render(connections_json)
    with open(profile_path, "w") as profiles:
        yaml.dump(yaml.safe_load(profiles_body), profiles, default_flow_style=False, allow_unicode=True)
    return
